# coding: utf-8

# TODO: check full mapping of one enum to another (is every value of mapped enum has pair and vice versa)
# TODO: pep8
# TODO: pylint
# TODO: generate docs
# TODO: rewrite exceptions texts & rename exception classes
import random

from rels import exceptions

class Column(object):
    __slots__ = ('_creation_order', 'primary', 'unique', 'single_type', 'name', 'index_name', 'no_index', 'related_name', 'external', 'primary_checks')

    _creation_counter = 0

    def __init__(self,
                 name=None,
                 unique=True,
                 primary=False,
                 external=False,
                 single_type=True,
                 index_name=None,
                 no_index=True,
                 related_name=None,
                 primary_checks=False):
        '''
        name usually setupped by Relation class. In constructor it used in tests
        '''
        self._creation_order = self.__class__._creation_counter
        self.__class__._creation_counter += 1

        self.primary = bool(primary)
        self.unique = bool(unique)

        self.single_type = single_type
        self.name = name
        self.index_name = index_name
        self.no_index = no_index
        self.related_name = related_name
        self.external = external
        self.primary_checks = primary_checks

    def __repr__(self):
        repr_str = 'Column(name=%(name)r, unique=%(unique)r, primary=%(primary)r, '\
            'single_type=%(single_type)r, index_name=%(index_name)r, related_name=%(related_name)r)'
        return repr_str % self.__dict__

    def initialize(self, name):
        self.name = name

        if self.index_name is None:
            self.index_name = 'index_%s' % self.name

        if self.primary and not self.unique:
            raise exceptions.PrimaryWithoutUniqueError(self.name)

        if self.external and not self.unique:
            raise exceptions.ExternalWithoutUniqueError(self.name)

    def check_uniqueness_restriction(self, records):
        if not self.unique: return

        values = set()

        for record in records:
            value = getattr(record, self.name)

            if value in values:
                raise exceptions.DuplicateValueError(self.name, value)

            values.add(value)

    def check_single_type_restriction(self, records):
        if not self.single_type: return

        if not records: return

        expected_type = getattr(records[0], self.name).__class__

        for record in records:
            value_type = getattr(record, self.name).__class__
            if expected_type != value_type:
                raise exceptions.SingleTypeError(self.name)

    def get_primary_attributes(self, records):
        if not records: return {}
        return { getattr(record, self.name):record for record in records}

    def get_index(self, records):

        index = {}

        if self.unique:
            for record in records:
                index[getattr(record, self.name)] = record
        else:
            for record in records:
                value = getattr(record, self.name)
                # save declaration order
                index[value] = index.get(value, []) + [record]

            index = { k:tuple(v) for k, v in index.items()}

        return index


class Record(object):

    def __init__(self, columns, data, relation_class=None):
        self._primaries = []
        self._relation = relation_class

        if len(columns) != len(data):
            raise exceptions.ColumnsNumberError(columns, data)

        for column, value in zip(columns, data):
            setattr(self, column.name, value)

            if column.related_name is not None:
                if not hasattr(value, 'set_related_name'):
                    raise exceptions.SetRelatedNameError(value)
                value.set_related_name(column.related_name, self)

    def __getattr__(self, name):
        if name.startswith('is_'):
            return getattr(self._relation, name[3:]) is self

        return getattr(super(), name)

    def _add_primary(self, primary_name):
        self._primaries.append(primary_name)

    def set_related_name(self, name, record):
        if hasattr(self, name):
            raise exceptions.DuplicateRelatonNameError(record, name)
        setattr(self, name, record)

    def _set_primary_checks(self, column, ids):
        if not column.primary_checks:
            return

        for id_ in ids:
            attr_name = 'is_%s' % id_
            if hasattr(self, attr_name):
                raise exceptions.DuplicateIsPrimaryError(self, column, attr_name, id_)
            setattr(self, attr_name, id_ == getattr(self, column.name))

    def __repr__(self):
        relation_name = self._relation.__name__ if self._relation is not None else None
        primary_name = self._primaries[0] if self._primaries else None
        return '%(relation)s.%(primary)s' % {'relation': relation_name,
                                             'primary': primary_name}

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self


class _RelationMetaclass(type):

    @classmethod
    def process_class_attributes(cls, relation_class, bases, attributes):
        relation_attributes = {}
        columns = {}
        raw_records = []

        for attr_name, attr_value in attributes.items():
            if attr_name == 'records':
                raw_records = attr_value
            elif isinstance(attr_value, Column):
                attr_value.initialize(name=attr_name)
                columns[attr_name] = attr_value
            else:
                relation_attributes[attr_name] = attr_value

        for base in bases:
            if hasattr(base, '_columns'):
                for column in base._columns:
                    if column.name not in columns:
                        columns[column.name] = column
            if hasattr(base, '_raw_records'):
                raw_records = list(base._raw_records) + list(raw_records)

        columns = sorted(columns.values(), key=lambda c: c._creation_order)

        external_columns = [column for column in columns if column.external]
        if len(external_columns) > 1:
            raise exceptions.MultipleExternalColumnsError(external_columns)

        records = [Record(columns, record, relation_class) for record in raw_records]

        relation_attributes['records'] = tuple(records)
        relation_attributes['_raw_records'] = tuple(raw_records)
        relation_attributes['_columns'] = columns
        relation_attributes['_external_index'] = {}

        return columns, relation_attributes, records

    def __new__(cls, name, bases, attributes):

        relation_class = super(_RelationMetaclass, cls).__new__(cls, name, bases, {})

        columns, relation_attributes, records = cls.process_class_attributes(relation_class,
                                                                          bases,
                                                                          attributes)

        for column in columns:
            column.check_uniqueness_restriction(records)
            column.check_single_type_restriction(records)

        # create primaries
        for column in columns:
            if not column.primary:
                continue
            attributes = column.get_primary_attributes(records)

            duplicates = list(set(attributes.keys()) & set(relation_attributes.keys()))
            if duplicates:
                raise exceptions.PrimaryDuplicatesRelationAttributeError(duplicates, column.name)

            for record in records:
                record._set_primary_checks(column, list(attributes.keys()))

            for attr_name, record in attributes.items():
                record._add_primary(attr_name)

            relation_attributes.update(attributes)

        # create indexes
        for column in columns:
            index = None

            if not column.no_index or column.external:
                index = column.get_index(records)

                if column.index_name in relation_attributes:
                    raise exceptions.IndexDuplicatesRelationAttributeError(column.name, column.index_name)

                relation_attributes[column.index_name] = index

                if column.external:
                    relation_attributes['_external_index'] = index

        for attr_name, attr_value in relation_attributes.items():
            setattr(relation_class, attr_name, attr_value)

        return relation_class

    def __call__(self, id_):
        if id_ not in self._external_index:
            raise exceptions.NotExternalValueError(id_)
        return self._external_index[id_]

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self



class Relation(object, metaclass=_RelationMetaclass):

    @classmethod
    def select(cls, *field_names):
        result = []
        for record in cls.records:
            row = tuple(getattr(record, field_name) for field_name in field_names)
            result.append(row)
        return tuple(result)

    @classmethod
    def random(cls, exclude=()):
        return random.choice([record for record in cls.records if record not in exclude])

    @classmethod
    def get_from_name(cls, name):
        # TODO: write tests
        relation_name, primary_name = name.split('.')

        if relation_name != cls.__name__:
            raise exceptions.WrongRelationNameError(relation_name=relation_name, enum_name=cls.__name__)

        return getattr(cls, primary_name)

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self



# codings: utf-8

from rels.relations import Column, Relation


class Enum(Relation):
    name = Column(primary=True, no_index=True, primary_checks=True)
    value = Column(external=True, no_index=True)


class EnumWithText(Relation):
    name = Column(primary=True, no_index=True, primary_checks=True)
    value = Column(external=True, no_index=True)
    text = Column()


class NullObject(object):

    def set_related_name(self, *argv, **kwargs):
        pass



# coding: utf-8


from django.db import models
from django.utils import functional
from django.core.exceptions import ValidationError
from django.core import validators as django_validators

from rels.relations import Record
from rels.shortcuts import EnumWithText


class DjangoEnum(EnumWithText):

    @classmethod
    def choices(cls):
        return [(record, record.text) for record in cls.records]


class RelationIntegerField(models.IntegerField):

    def __init__(self, *argv, **kwargs):
        self._relation = kwargs.get('relation')
        self._relation_column = kwargs.get('relation_column', 'value')

        if 'choices' not in kwargs and hasattr(self._relation, 'choices'):
            kwargs['choices'] = self._relation.choices()

        # TODO: check if relation has column with relation_column name
        # TODO: check if column has int type

        if 'relation' in kwargs: del kwargs['relation']
        if 'relation_column' in kwargs: del kwargs['relation_column']

        super(RelationIntegerField, self).__init__(*argv, **kwargs)

    def to_python(self, value):
        if self._relation is None:
            # emulate default behaviour for migrations
            return super(RelationIntegerField, self).to_python(value)

        if value is None:
            return None

        if isinstance(value, Record):
            if value._relation == self._relation:
                return value
            else:
                raise ValidationError('record %r is not from %r' % (value, self._relation))

        if isinstance(value, str) and '.' in value:
            relation_name, primary_name = value.split('.')

            if relation_name != self._relation.__name__:
                raise ValidationError('wrong relation name "%s", expected "%s"' % (relation_name, self._relation.__name__))

            return getattr(self._relation, primary_name)

        try:
            return getattr(self._relation, 'index_%s' % self._relation_column)[int(value)]
        except ValueError:
            raise ValidationError('can not convert %r to %r' % (value, self._relation))

    def get_prep_value(self, value):
        if self._relation is None:
            # emulate default behaviour for migrations
            return super(RelationIntegerField, self).get_prep_value(value)

        if isinstance(value, Record):
            if value._relation == self._relation:
                return getattr(value, self._relation_column)
            else:
                # TODO: change exception type
                raise ValidationError('record %r is not from %r' % (value, self._relation))

        if isinstance(value, str) and '.' in value:
            relation_name, primary_name = value.split('.')

            if relation_name != self._relation.__name__:
                # TODO: change exception type
                raise ValidationError('wrong relation name "%s", expected "%s"' % (relation_name, self._relation.__name__))

            return getattr(getattr(self._relation, primary_name), self._relation_column)

        return value

    def from_db_value(self, value, expression, connection):
        return self.to_python(value)

    def deconstruct(self):
        name, path, args, kwargs = super(RelationIntegerField, self).deconstruct()
        if 'choices' in kwargs:
            del kwargs['choices']
        return name, path, args, kwargs

    @functional.cached_property
    def validators(self):
        validators = super(RelationIntegerField, self).validators

        # remove unnecessary validators
        validators = [validator
                      for validator in validators
                      if not isinstance(validator, (django_validators.MinValueValidator, django_validators.MaxValueValidator))]

        return validators



# coding: utf-8

########################
# Базовое использование
########################

from rels import Column, Relation

# Enum и EnumWithText уже объявлены в библиотеке
# и доступны как rels.Enum и rels.EnumWithText
# тут их объявления привидены для упрощения понимания

class Enum(Relation):             # объявляем абстраткное перечисление
    name = Column(primary=True)   # имя
    value = Column(external=True) # значение


# наследование — добавляем дополнительный столбец для какого-нибудь текста
#                например, для использования в пользовательском интерфейсе
class EnumWithText(Enum):
    text = Column()


class SOME_CONSTANTS(Enum):       # объявляем конкретное перечисление
    records = ( ('NAME_1', 1),    # и указываем данные для него
                ('NAME_2', 2))


class SOME_CONSTANTS_WITH_TEXT(EnumWithText): # ещё одно конкретное перечисление
    records = ( ('NAME_1', 1, 'constant 1'),
                ('NAME_2', 2, 'constant 2'))


# Работаем с перечислениями

# доступ к данным
SOME_CONSTANTS.NAME_1.name == 'NAME_1'          # True
SOME_CONSTANTS.NAME_1.value == 1                # True

# получение элемента перечисления из «сырых» данных
SOME_CONSTANTS(1) == SOME_CONSTANTS.NAME_1      # True

# сравнения
SOME_CONSTANTS.NAME_2 == SOME_CONSTANTS.NAME_2  # True
SOME_CONSTANTS.NAME_2 != SOME_CONSTANTS.NAME_1  # True

# теперь для проверок не надо всюду тягать импорты перечисления
SOME_CONSTANTS.NAME_2.is_NAME_1                 # False
SOME_CONSTANTS.NAME_2.is_NAME_2                 # True

# каждый элемент перечисления — отдельный объект,
# поэтому даже объекты с одинаковыми данными равны не будут
SOME_CONSTANTS.NAME_2 != SOME_CONSTANTS_WITH_TEXT.NAME_2  # True
SOME_CONSTANTS.NAME_1 != SOME_CONSTANTS_WITH_TEXT.NAME_1  # True

# наследование — добавляем новые элементы
class EXTENDED_CONSTANTS(SOME_CONSTANTS_WITH_TEXT):  # расширяем набор данных в перечислении
    records = ( ('NAME_3', 3, 'constant 3'), )       # добавляем ещё одно значение


########################
# Индексы
########################

class ENUM(Relation):
    name = Column(primary=True)   # для этого столбца имя индекса будет .index_name
    value = Column(external=True) # для этого столбца имя индекса будет .index_value
    text = Column(unique=False, index_name='by_key') # указываем своё имя для индекса

    records = ( ('NAME_1', 0, 'key_1'),
                ('NAME_2', 1, 'key_2'),
                ('NAME_3', 2, 'key_2'), )

# если данные в столбце уникальны, значением в словаре будет элемент перечисления
ENUM.index_name # {'NAME_1': ENUM.NAME_1, 'NAME_2': ENUM.NAME_2,  'NAME_3': ENUM.NAME_3}

# если данные в столбце не уникальны, значением в словаре будет список элементов перечисления
ENUM.by_key     # {'key_1': [ENUM.NAME_1], 'key_2': [ENUM.NAME_2, ENUM.NAME_3]}


########################
# Обратные ссылки
########################

# объявляем отношение, на которое будем ссылаться
class DESTINATION_ENUM(Relation):
    name = Column(primary=True)
    val = Column()

    records = ( ('STATE_1', 'value_1'),
                ('STATE_2', 'value_2') )

# объявляем отношение, которое будет ссылаться
class SOURCE_ENUM(Relation):
    name = Column(primary=True)
    val = Column()
    rel = Column(related_name='rel_source')

    records = ( ('STATE_1', 'value_1', DESTINATION_ENUM.STATE_1),
                ('STATE_2', 'value_2', DESTINATION_ENUM.STATE_2) )

# проверяем работу ссылок
DESTINATION_ENUM.STATE_1.rel_source == SOURCE_ENUM.STATE_1 # True
DESTINATION_ENUM.STATE_2 == SOURCE_ENUM.STATE_2.rel        # True



# coding: utf-8
import copy

from unittest import TestCase

from rels.relations import Relation, Column, Record

from rels import exceptions

class Enum(Relation):
    name = Column(primary=True, no_index=False, primary_checks=True)
    value = Column(external=True, no_index=False)


class EnumWithText(Relation):
    name = Column(primary=True, no_index=False, primary_checks=True)
    value = Column(external=True, no_index=False)
    text = Column(no_index=False)


class EmptyRelation(Relation):
    pass

# just to test, that empy records does not break class costruction
class EmptyRecordsRelation(Relation):
    name = Column(no_index=False)
    value = Column(no_index=False)


class SimplestRelation(Relation):
    name = Column(no_index=False)
    value = Column(no_index=False)

    records = (('name_a', 1),
                ('name_b', 2))


class SimplestEnum(Relation):
    name = Column(primary=True, no_index=False, primary_checks=True)
    value = Column(external=True, no_index=False)

    records = ( ('state_1', 'val_1'),
                 ('state_2', 'val_2') )

class EnumWith2Primaries(Relation):

    name = Column(primary=True, no_index=False, primary_checks=True)
    value = Column(primary=True, no_index=False, primary_checks=True)

    records = ( ('STATE_1', 'value_1'),
                 ('STATE_2', 'value_2') )

class IndexesRelation(Relation):
    name = Column(primary=True, no_index=False, primary_checks=True)
    val_1 = Column(no_index=False)
    val_2 = Column(unique=False, no_index=False)
    val_3 = Column(index_name='val_3_index', no_index=False)

    records = ( ('rec_1', 'val_1_1', 'val_2_1', 'val_3_1'),
                 ('rec_2', 'val_1_2', 'val_2_2', 'val_3_2'),
                 ('rec_3', 'val_1_3', 'val_2_2', 'val_3_3'),
                 ('rec_4', 'val_1_4', 'val_2_4', 'val_3_4'),)

class RelationDestinationRelation(Relation):
    name = Column(primary=True, no_index=False, primary_checks=True)
    val_1 = Column(no_index=False)

    records = ( ('STATE_1', 'value_1'),
                 ('STATE_2', 'value_2') )

class RelationSourceRelation(Relation):
    name = Column(primary=True, no_index=False, primary_checks=True)
    val_1 = Column(no_index=False)
    rel = Column(related_name='rel_source', no_index=False)

    records = ( ('STATE_1', 'value_1', RelationDestinationRelation.STATE_1),
                 ('STATE_2', 'value_2', RelationDestinationRelation.STATE_2) )

class ShortcutEnum(Enum):
    records = ( ('ID_1', 1),
                 ('ID_2', 2) )

class ShortcutEnumWithText(EnumWithText):
    records = ( ('ID_1', 1, 'verbose name 1'),
                 ('ID_2', 2, 'verbose name 2') )



class SimpleColumnTests(TestCase):

    def test_default_index_name(self):
        column = Column()
        column.initialize(name='test')
        self.assertEqual(column.index_name, 'index_test')

    def test_custom_index_name(self):
        column = Column(index_name='my_index')
        column.initialize(name='test')
        self.assertEqual(column.index_name, 'my_index')

    def test_check_uniqueness_restriction_success(self):
        column_1 = Column()
        column_1.initialize('column_1')

        records = ( Record([column_1], ['uuid_1']),
                    Record([column_1], ['uuid_2']))

        column_1.check_uniqueness_restriction(records)

    def test_primary_without_unique(self):
        column = Column(primary=True, unique=False)
        self.assertRaises(exceptions.PrimaryWithoutUniqueError, column.initialize, name='column')

    def test_external_without_unique(self):
        column = Column(external=True, unique=False)
        self.assertRaises(exceptions.ExternalWithoutUniqueError, column.initialize, name='column')

    def test_check_uniqueness_restriction_exception(self):
        column_1 = Column()
        column_1.initialize('column_1')

        records = ( Record([column_1], ['uuid_1']),
                    Record([column_1], ['uuid_1']))

        self.assertRaises(exceptions.DuplicateValueError,
                          column_1.check_uniqueness_restriction, records)

    def test_check_single_type_restriction_success(self):
        column_1 = Column()
        column_1.initialize('column_1')

        records = ( Record([column_1], ['str_1']),
                    Record([column_1], ['str_2']))

        column_1.check_single_type_restriction(records)

    def test_check_single_type_restriction_exception(self):
        column_1 = Column()
        column_1.initialize('column_1')

        records = ( Record([column_1], [1]),
                    Record([column_1], ['str_2']))

        self.assertRaises(exceptions.SingleTypeError,
                          column_1.check_single_type_restriction, records)

    # def test_get_index__no_index(self):
    #     column_1 = Column(no_index=True)
    #     column_1.initialize('column_1')

    #     records = ( Record([column_1], ['str_1']),
    #                 Record([column_1], ['str_2']))

    #     self.assertEqual(column_1.get_index(records), None)

    def test_get_index_unique(self):
        column_1 = Column()
        column_1.initialize('column_1')

        records = ( Record([column_1], ['str_1']),
                    Record([column_1], ['str_2']))

        self.assertEqual(column_1.get_index(records), {'str_1': records[0],
                                                       'str_2': records[1]})

    def test_get_index_not_unique(self):
        column_1 = Column(unique=False)
        column_1.initialize('column_1')

        records = ( Record([column_1], ['str_1']),
                    Record([column_1], ['str_2']))

        self.assertEqual(column_1.get_index(records), {'str_1': (records[0], ),
                                                       'str_2': (records[1], )})

    # def test_repr(self):
    #     column = Column(related_name='rel_name')
    #     column.initialize(name='col_name')

    #     column._creation_order += 1 # enshure, that creation order will be equal
    #     self.assertEqual(eval(repr(column)).__dict__, column.__dict__)


class SimpleRecordTests(TestCase):

    def test_wrong_data_length(self):
        columns = (Column(name='col_1'), Column(name='col_2'))
        data = (1, 2, 3)
        self.assertRaises(exceptions.ColumnsNumberError, Record, columns=columns, data=data)

    def test_columns_values(self):
        columns = (Column(name='col_1'), Column(name='col_2'))
        data = (1, 2)
        r = Record(columns=columns, data=data)
        self.assertEqual(r.col_1, 1)
        self.assertEqual(r.col_2, 2)

    def test_equality(self):
        columns = (Column(name='col_1'), Column(name='col_2'))
        data = (1, 2)
        r_1 = Record(columns=columns, data=data)
        r_2 = Record(columns=columns, data=data)
        self.assertNotEqual(r_1, r_2)
        self.assertEqual(r_1, r_1)

    def test_repr(self):
        self.assertEqual(repr(RelationDestinationRelation.records[0]),
                         'RelationDestinationRelation.STATE_1')


class SimpleRelationTests(TestCase):

    def setUp(self):
        pass

    def test_empy_relation(self):
        self.assertEqual(EmptyRelation.records, tuple())

    def test_simplest_relation(self):
        self.assertEqual(len(SimplestRelation.records), 2)
        self.assertEqual(SimplestRelation._raw_records, (('name_a', 1),
                                                      ('name_b', 2)))

    def test_uniqueness_restriction(self):
        def create_bad_relation():
            class SimplestRelation(Relation):
                name = Column()
                value = Column(unique=True)

                records = (('name_a', 1),
                            ('name_b', 1))

        self.assertRaises(exceptions.DuplicateValueError, create_bad_relation)

    def test_primary_uniqueness_restriction(self):
        def create_bad_relation():
            class SimplestRelation(Relation):
                name = Column(primary=True, unique=False)
                value = Column()

                records = (('name_a', 1),
                            ('name_b', 1))

        self.assertRaises(exceptions.PrimaryWithoutUniqueError, create_bad_relation)


    def test_simplest_enum_attributes(self):
        self.assertEqual(SimplestEnum.state_1.value, 'val_1')
        self.assertEqual(SimplestEnum.state_1.name, 'state_1')
        self.assertEqual(SimplestEnum.state_2.value, 'val_2')
        self.assertEqual(SimplestEnum.state_2.name, 'state_2')

    def test_primary_attributes_setupped(self):
        self.assertEqual(SimplestEnum.records, (SimplestEnum.state_1, SimplestEnum.state_2))

    def test_only_primary_attributes_setupped(self):
        self.assertFalse(hasattr(SimplestEnum, 'val_1'))
        self.assertFalse(hasattr(SimplestEnum, 'val_2'))

    def test_records_check_attributes_setupped(self):
        self.assertTrue(SimplestEnum.state_1.is_state_1)
        self.assertFalse(SimplestEnum.state_1.is_state_2)
        self.assertFalse(SimplestEnum.state_2.is_state_1)
        self.assertTrue(SimplestEnum.state_2.is_state_2)

        self.assertFalse(hasattr(SimplestEnum.state_1, '_is_val_1'))
        self.assertFalse(hasattr(SimplestEnum.state_1, '_is_val_2'))
        self.assertFalse(hasattr(SimplestEnum.state_2, '_is_val_1'))
        self.assertFalse(hasattr(SimplestEnum.state_2, '_is_val_2'))

    def test_get_record_by_not_external_id(self):
        self.assertRaises(exceptions.NotExternalValueError, SimplestEnum, 'bla-bla')

    def test_more_then_1_external_columns(self):
        def create_bad_relation():
            class SimplestRelation(Relation):
                name = Column(external=True)
                value = Column(external=True)

                records = (('name_a', 'name_c'),
                            ('name_b', 'name_d'))

        self.assertRaises(exceptions.MultipleExternalColumnsError, create_bad_relation)

    def test_get_record_by_external_id(self):
        self.assertEqual(SimplestEnum('val_1'), SimplestEnum.state_1)
        self.assertEqual(SimplestEnum('val_2'), SimplestEnum.state_2)

    def test_simplest_enum_attributes_with_2_primaries(self):
        self.assertEqual(EnumWith2Primaries.STATE_1.value, 'value_1')
        self.assertEqual(EnumWith2Primaries.STATE_1.name, 'STATE_1')
        self.assertEqual(EnumWith2Primaries.STATE_2.value, 'value_2')
        self.assertEqual(EnumWith2Primaries.STATE_2.name, 'STATE_2')

        self.assertEqual(EnumWith2Primaries.STATE_1,
                         EnumWith2Primaries.value_1)

        self.assertEqual(EnumWith2Primaries.STATE_2,
                         EnumWith2Primaries.value_2)

        self.assertEqual(EnumWith2Primaries.records,
                         (EnumWith2Primaries.STATE_1,
                          EnumWith2Primaries.STATE_2))

    def test_primary_name_duplicate_another_relation_attribute(self):
        def create_bad_relation():
            class SimplestRelation(Relation):
                name = Column(primary=True)
                value = Column(primary=True)

                records = (('name_a', 'name_c'),
                            ('name_b', 'name_a'))

        self.assertRaises(exceptions.PrimaryDuplicatesRelationAttributeError, create_bad_relation)

    def test_primary_name_duplicate_another_record_attribute(self):
        def create_bad_relation():
            class SimplestRelation(Relation):
                name = Column(primary=True, no_index=False, primary_checks=True)
                is_name = Column(primary=True, no_index=False, primary_checks=True)

                records = (('name_a', 'name_c'),
                           ('name_b', 'name'))

        self.assertRaises(exceptions.DuplicateIsPrimaryError, create_bad_relation)


    def test_simplest_indexes(self):
        self.assertEqual(IndexesRelation.index_name,
                         {'rec_1': IndexesRelation.rec_1,
                          'rec_2': IndexesRelation.rec_2,
                          'rec_3': IndexesRelation.rec_3,
                          'rec_4': IndexesRelation.rec_4 })

        self.assertEqual(IndexesRelation.index_val_1,
                         {'val_1_1': IndexesRelation.rec_1,
                          'val_1_2': IndexesRelation.rec_2,
                          'val_1_3': IndexesRelation.rec_3,
                          'val_1_4': IndexesRelation.rec_4 })

        self.assertEqual(IndexesRelation.index_val_2,
                         {'val_2_1': (IndexesRelation.rec_1,),
                          'val_2_2': (IndexesRelation.rec_2, IndexesRelation.rec_3),
                          'val_2_4': (IndexesRelation.rec_4,) })

        self.assertEqual(IndexesRelation.val_3_index,
                         {'val_3_1': IndexesRelation.rec_1,
                          'val_3_2': IndexesRelation.rec_2,
                          'val_3_3': IndexesRelation.rec_3,
                          'val_3_4': IndexesRelation.rec_4 })

    def test_index_duplicate_another_relation_attribute(self):
        def create_bad_relation():
            class SimplestRelation(Relation):
                name = Column(primary=True, index_name='name_a', no_index=False)
                value = Column(no_index=False)

                records = (('name_a', 1),
                            ('name_b', 2))

        self.assertRaises(exceptions.IndexDuplicatesRelationAttributeError, create_bad_relation)

    def test_relations_setup(self):
        self.assertEqual(RelationSourceRelation.STATE_1, RelationDestinationRelation.STATE_1.rel_source)
        self.assertEqual(RelationSourceRelation.STATE_2, RelationDestinationRelation.STATE_2.rel_source)
        self.assertEqual(RelationSourceRelation.STATE_1.rel, RelationDestinationRelation.STATE_1)
        self.assertEqual(RelationSourceRelation.STATE_2.rel, RelationDestinationRelation.STATE_2)
        self.assertEqual(RelationSourceRelation.STATE_1, RelationSourceRelation.STATE_1.rel.rel_source)
        self.assertEqual(RelationSourceRelation.STATE_2, RelationSourceRelation.STATE_2.rel.rel_source)

    def test_relations_setup_without_set_method(self):
        def create_bad_relation():
            class RelationSourceRelation(Relation):
                name = Column(primary=True)
                val_1 = Column()
                rel = Column(related_name='rel_source')

                records = ( ('STATE_1', 'value_1', 1), # just any type without .set_related_name
                             ('STATE_2', 'value_2', RelationDestinationRelation.STATE_2) )

        self.assertRaises(exceptions.SetRelatedNameError, create_bad_relation)

    def test_relations_setup_duplicate_name_error(self):
        def create_bad_relation():
            class RelationSourceRelation(Relation):
                name = Column(primary=True)
                val_1 = Column()
                rel = Column(related_name='rel_source')

                # "name" duplicate primary attribute of RelationDestinationRelation
                records = ( ('name', 'value_1', RelationDestinationRelation.STATE_1),
                             ('STATE_2', 'value_2', RelationDestinationRelation.STATE_2) )

        self.assertRaises(exceptions.DuplicateRelatonNameError, create_bad_relation)

    def test_relation_inheritance(self):
        class BaseRelation(Relation):
            name = Column(primary=True)
            value = Column()

        class ChildRelation(BaseRelation):
            records = (('id_1', 1),
                        ('id_2', 2))

        self.assertEqual(ChildRelation.id_1.name, 'id_1')
        self.assertEqual(ChildRelation.id_2.name, 'id_2')

    def test_column_redifinition(self):
        class BaseRelation(Relation):
            name = Column(primary=True, no_index=False)
            value = Column(no_index=False)

        class ChildRelation(BaseRelation):
            value = Column(unique=False, no_index=False)
            records = (('id_1', 1),
                        ('id_2', 1))

        self.assertEqual(ChildRelation.index_value, {1: ChildRelation.records})

    def test_relation_inheritance_with_records(self):
        class BaseRelation(Relation):
            name = Column(primary=True, no_index=False)
            value = Column(no_index=False)
            records = (('id_0', 0),)

        class ChildRelation(BaseRelation):
            value = Column(unique=False, no_index=False)
            records = (('id_1', 1),
                        ('id_2', 2))

        self.assertEqual(ChildRelation.id_0.name, 'id_0')
        self.assertEqual(ChildRelation.id_1.name, 'id_1')
        self.assertEqual(ChildRelation.id_2.name, 'id_2')

        self.assertEqual(len(ChildRelation.records), 3)
        self.assertEqual(ChildRelation._raw_records, (('id_0', 0),
                                                   ('id_1', 1),
                                                   ('id_2', 2),))

        self.assertRaises(AttributeError, getattr, BaseRelation, 'id_1')

        self.assertEqual(len(BaseRelation.index_name), 1)
        self.assertEqual(len(ChildRelation.index_name), 3)


    def test_shortcut_enum(self):
        self.assertEqual(ShortcutEnum.ID_1.value, 1)
        self.assertEqual(ShortcutEnum.ID_2.name, 'ID_2')

    def test_shortcur_enum_with_text(self):
        self.assertEqual(ShortcutEnumWithText.ID_1.value, 1)
        self.assertEqual(ShortcutEnumWithText.ID_2.name, 'ID_2')
        self.assertEqual(ShortcutEnumWithText.ID_2.text, 'verbose name 2')

    def test_hash(self):
        container = set()
        container.add(SimplestEnum.state_1)
        container.add(SimplestEnum.state_2)
        self.assertEqual(len(container), 2)

        container.add(SimplestEnum.state_2)
        self.assertEqual(len(container), 2)

        container = {SimplestEnum.state_2: 666}
        self.assertEqual(container[SimplestEnum.state_2], 666)

    def test_copy_relation(self):
        self.assertEqual(id(SimplestRelation), id(copy.copy(SimplestRelation)))

    def test_deepcopy_relation(self):
        self.assertEqual(id(SimplestRelation), id(copy.deepcopy(SimplestRelation)))

    def test_copy_record(self):
        self.assertEqual(id(SimplestRelation.records[0]), id(copy.copy(SimplestRelation.records[0])))

    def test_deepcopy_record(self):
        self.assertEqual(id(SimplestRelation.records[0]), id(copy.deepcopy(SimplestRelation.records[0])))



# coding: utf-8

class RelsException(Exception): pass
class ColumnException(RelsException): pass
class RecordException(RelsException): pass
class RelationException(RelsException): pass

class PrimaryWithoutUniqueError(ColumnException):
    def __init__(self, column_name):
        message = 'Primary column "%s" MUST has unique restriction' % column_name
        super(PrimaryWithoutUniqueError, self).__init__(message)

class ExternalWithoutUniqueError(ColumnException):
    def __init__(self, column_name):
        message = 'External column "%s" MUST has unique restriction' % column_name
        super(ExternalWithoutUniqueError, self).__init__(message)

class DuplicateValueError(ColumnException):
    def __init__(self, column_name, value):
        message = 'Duplicate value "%s" in column "%s"' % (value, column_name)
        super(DuplicateValueError, self).__init__(message)

class SingleTypeError(ColumnException):
    def __init__(self, column_name):
        message = 'Column "%s" MUST contain values of one type' % (column_name)
        super(SingleTypeError, self).__init__(message)

class ColumnsNumberError(RecordException):
    def __init__(self, columns, data):
        message = 'Wrong columns number in record: %s (expected: %d)' % (data, len(columns))
        super(ColumnsNumberError, self).__init__(message)

class SetRelatedNameError(RecordException):
    def __init__(self, object):
        message = ('Can not set related name for object %r, it has no method "set_related_name"'
                   % object)
        super(SetRelatedNameError, self).__init__(message)

class DuplicateRelatonNameError(RecordException):
    def __init__(self, record, name):
        message = 'Can not set related name for record %s, it has already had attribute with name "%s"' % (record, name)
        super(DuplicateRelatonNameError, self).__init__(message)

class DuplicateIsPrimaryError(RecordException):
    def __init__(self, record, column, attr_name, primary_name):
        message = ('record %(record)s attribute %(attr_name)s of column %(column_name)s duplicates another record attribute (probably, column with name "%(primary_name)s")' %
        {'attr_name': attr_name,
         'record': record,
         'column_name': column.name,
         'primary_name': primary_name})
        super(DuplicateIsPrimaryError, self).__init__(message)

class PrimaryDuplicatesRelationAttributeError(RelationException):
    def __init__(self, column_name, duplicates):
        message = 'Primary names "%(duplicates)r" of column "%(column)s" duplicate another table attributes' % {'duplicates': duplicates, 'column': column_name}
        super(PrimaryDuplicatesRelationAttributeError, self).__init__(message)

class IndexDuplicatesRelationAttributeError(RelationException):
    def __init__(self, column_name, index_name):
        message = ('Index name "%s" of column "%s" duplicates another table attribute'
                   % (index_name, column_name))
        super(IndexDuplicatesRelationAttributeError, self).__init__(message)

class NotExternalValueError(RelationException):
    def __init__(self, id_):
        message = '"%(id)s" is not external value' % {'id': id_}
        super(NotExternalValueError, self).__init__(message)

class MultipleExternalColumnsError(RelationException):
    def __init__(self, external_columns):
        message = ('there are more then 1 external column: %s' %
                   ', '.join(column.name for column in external_columns))
        super(MultipleExternalColumnsError, self).__init__(message)


class WrongRelationNameError(RelationException):

    def __init__(self, relation_name, enum_name):
        message = 'wrong relation name "%s", expected enum name: "%s"' % (relation_name, enum_name)
        super(WrongRelationNameError, self).__init__(message)



# coding: utf-8
import setuptools

setuptools.setup(
    name='Rels',
    version='0.3.1',
    description='Library for describing data relations in python (Enums, "relational" tables)',
    long_description = open('README.rst').read(),
    url='https://github.com/Tiendil/rels',
    author='Aleksey Yeletsky <Tiendil>',
    author_email='a.eletsky@gmail.com',
    license='BSD',
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Developers',

        'Topic :: Software Development :: Libraries :: Python Modules',

        'License :: OSI Approved :: BSD License',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',

        'Natural Language :: English',
        'Natural Language :: Russian'],
    keywords=['enums', 'data relations', 'relational tables'],
    packages=setuptools.find_packages(),
    include_package_data=True,
    test_suite = 'tests',
    )



# coding: utf-8

from rels.relations import Column, Record, Relation
from rels import exceptions
from .shortcuts import Enum, EnumWithText, NullObject

__all__ = [Column, Record, Relation, exceptions, Enum, EnumWithText, NullObject]
